# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import subprocess

from ubuntu_system_tests.helpers.processes import is_process_running
from ubuntu_system_tests.helpers.testbed import run_command_with_sudo
from ubuntu_system_tests.helpers.utils import wait_until


def run_command(command, sudo=False, *args, **kwargs):
    """Run the provided command
    :param command: The command to run
    :param sudo: True if you want to run the command with sudo privileges
    :param args: List with the arguments for the command
    :return: Output of command.
    """
    kwargs['universal_newlines'] = kwargs.get('universal_newlines', True)
    if sudo:
        return run_command_with_sudo(command, *args, **kwargs)
    else:
        command = '{c} {a}'.format(c=command, a=' '.join(args)).strip()
        return subprocess.check_output(command, shell=True,
                                       stderr=subprocess.STDOUT, **kwargs)


def is_systemd_running():
    """ Indicate if systemd process is running """
    return is_process_running('systemd')


class Job:
    """This contains a base implementation for a Job entity and its
     common functions like:
     start, stop, get_status, etc.
     """

    def __init__(self, name, sudo=False):
        if is_systemd_running():
            self.job = JobSystemD(name, sudo)
        else:
            self.job = JobUpstart(name, sudo)

    def start(self, *args, **kwargs):
        self.job.start(*args, **kwargs)

    def stop(self, *args, **kwargs):
        self.job.stop(*args, **kwargs)

    def restart(self, *args, **kwargs):
        self.job.restart(*args, **kwargs)

    def get_pid(self, *args, **kwargs):
        return self.job.get_pid(*args, **kwargs)

    def is_running(self, *args, **kwargs):
        return self.job.is_running(*args, **kwargs)

    def wait_for_job_spawned(self, timeout=10):
        self.job.wait_for_job_spawned(timeout)

    def ensure_stopped(self, timeout=300, *args, **kwargs):
        self.job.ensure_stopped(timeout, *args, **kwargs)


class JobBase:
    """ This contains a base implementation for a Job entity """

    def __init__(self, name, sudo):
        self.name = name
        self.sudo = sudo
        self.start_command = None
        self.stop_command = None
        self.restart_command = None

    def start(self, *args, **kwargs):
        """Start the job
        :param args: The arguments to be used when starting the job.
        """
        run_command(self.start_command, self.sudo, *args, **kwargs)

    def stop(self, *args, **kwargs):
        """Stop the job"""
        run_command(self.stop_command, self.sudo, *args, **kwargs)

    def restart(self, *args, **kwargs):
        """Restart the job"""
        run_command(self.restart_command, self.sudo, *args, **kwargs)

    def is_running(self, *args, **kwargs):
        pass

    def ensure_stopped(self, timeout=300, *args, **kwargs):
        """Stops the job if it is running and then checks
        that it effectively was stopped
        """
        if self.is_running(*args, **kwargs):
            self.stop()
        if not wait_until(lambda: not self.is_running(), timeout=timeout):
            raise RuntimeError(
                'Job {} not stopped within timeout {}'.format(
                    self.name, timeout))


class JobUpstart(JobBase):
    """This class contains the implementation for upstart managed jobs """

    def __init__(self, name, sudo=False):
        super().__init__(name, sudo)
        self.start_command = '/sbin/initctl start {}'.format(name)
        self.stop_command = '/sbin/initctl stop {}'.format(name)
        self.status_command = '/sbin/initctl status {}'.format(name)
        self.restart_command = '/sbin/initctl restart {}'.format(name)

    def _get_status(self, *args, **kwargs):
        """Gets the current status of the job"""
        return run_command(self.status_command, self.sudo, *args,
                           **kwargs)

    def get_pid(self, *args, **kwargs):
        """Gets the current pid of the running job"""
        status = self._get_status(*args, **kwargs)
        if "start/" not in status:
            raise ValueError(
                '{} is not in the running state.'.format(self.name))
        pid = status.split()[-1]
        if not pid.isdigit():
            raise ValueError(
                '{} does not have valid process id: {}'.format(self.name, pid))
        return int(pid)

    def is_running(self, *args, **kwargs):
        """Return True if the job is running. Otherwise, False."""
        return 'start/' in self._get_status(*args, **kwargs)

    def wait_for_job_spawned(self, timeout=10):
        """Wait until the named job is in spawned state."""
        spawned = wait_until(
            lambda: 'start/spawned' in self._get_status(), timeout=timeout)
        if not spawned:
            raise RuntimeError(
                'Job {} not in start/spwaned state within timeout {}'.format(
                    self.name, timeout))


class JobSystemD(JobBase):
    """This class contains the implementation for systemd managed jobs """

    def __init__(self, name, sudo=False):
        super().__init__(name, sudo)
        self.start_command = 'systemctl start {}'.format(name)
        self.stop_command = 'systemctl stop {}'.format(name)
        self.restart_command = 'systemctl restart {}'.format(name)
        self.active_command = 'systemctl show {} -p ActiveState'.format(name)
        self.show_command = 'systemctl show {} -p ExecMainPID'.format(name)

    def _get_value_from_show_output(self, line, key):
        line = line.replace('\n', '')
        if line.startswith(key):
            return line.split('=')[1].strip()
        else:
            raise ValueError(
                'Job {} does not have valid key {} in output {}'.format(
                    self.name, key, line))

    def get_pid(self, *args, **kwargs):
        """Gets the current pid of the running job"""
        if not self.is_running():
            raise ValueError(
                '{} is not in the running state.'.format(self.name))
        line = run_command(self.show_command, self.sudo, *args, **kwargs)
        pid = self._get_value_from_show_output(line, 'ExecMainPID')
        if not pid.isdigit():
            raise ValueError(
                '{} does not have valid process id: {}'.format(
                    self.name, pid))
        return int(pid)

    def is_running(self, *args, **kwargs):
        """Return True if the job is running. Otherwise, False."""
        line = run_command(self.active_command, self.sudo, *args, **kwargs)
        state = self._get_value_from_show_output(line, 'ActiveState')
        return 'active' == state

    def wait_for_job_spawned(self, timeout=10):
        """Wait until the named job is in spawned state."""
        spawned = wait_until(self.is_running, timeout=timeout)
        if not spawned:
            raise RuntimeError(
                'Job {} not in start/spwaned state within timeout {}'.format(
                    self.name, timeout))

    @staticmethod
    def _get_global_param(global_):
        if global_:
            return '--global'
        return ''

    @staticmethod
    def set_environment_variable(key, value, global_=False):
        """ Set a systemd environment variable
        :param key: the variable name
        :param value: the variable value
        :param global_: if the variable has to be defined as global
        """
        isglobal = JobSystemD._get_global_param(global_)
        set_command = 'systemctl set-environment {}={} {}'.format(key, value,
                                                                  isglobal)
        run_command(set_command, sudo=True)

    @staticmethod
    def unset_environment_variable(key, global_=False):
        """ Unset an environment variable for systemd
        :param key: the variable name
        :param global_: if the variable has been set as global
        """
        isglobal = JobSystemD._get_global_param(global_)
        unset_command = 'systemctl unset-environment {} {}'.format(key,
                                                                   isglobal)
        run_command(unset_command, sudo=True)

    @staticmethod
    def get_environment_variable(key, global_=False):
        """ Retrieve the environment variable value
        :param key: the variable name
        :param global_: if the variable has been set as global
        """
        isglobal = JobSystemD._get_global_param(global_)
        set_command = 'systemctl show-environment {}'.format(isglobal)
        vars = run_command(set_command, sudo=True)
        for var in vars.split('\n'):
            keyval = var.split('=')
            if len(keyval) == 2 and keyval[0] == key:
                return keyval[1]
        return None

    @staticmethod
    def is_environment_variable(key, global_=False):
        """ Check if a variable variable is set. When the value is an empty
        string, the methods returns True
        :param key: the variable name
        :param global_: if the variable has been set as global
        """
        return not JobSystemD.get_environment_variable(key, global_) is None
