# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from subprocess import CalledProcessError

import fixtures

from ubuntu_system_tests.helpers import sensors
from ubuntu_system_tests.helpers.power import RepowerdJob
from ubuntu_system_tests.helpers.power import utils as power
from ubuntu_system_tests.helpers.processes import get_process_id
from ubuntu_system_tests.helpers.sensors import (
    validate_fifo_ready,
    setup_sensors
)

GSETTINGS_BRIGHTNESS_MAX = 255


class BrightnessEnvironmentFixture(fixtures.Fixture):

    def __init__(self, auto_brightness='disable',
                 brightness_value=GSETTINGS_BRIGHTNESS_MAX):
        self.auto_brightness = auto_brightness
        self.brightness_value = str(brightness_value)

    def setUp(self):
        super().setUp()
        self.useFixture(AutoBrightness(state=self.auto_brightness))
        self.useFixture(Brightness(value=self.brightness_value))


class AutoBrightness(fixtures.Fixture):

    def __init__(self, state):
        self.request = state

    def setUp(self):
        super().setUp()
        old_value = power.get_auto_brightness_current_setting()
        self.addCleanup(power.set_auto_brightness_setting, old_value)

        if self.request == 'enable':
            power.set_auto_brightness_setting('true')
        elif self.request == 'disable':
            power.set_auto_brightness_setting('false')


class Brightness(fixtures.Fixture):
    """Fixture to set custom screen brightness."""

    def __init__(self, value):
        self.value = value

    def setUp(self):
        super().setUp()
        old_value = power.get_display_brightness()
        power.set_display_brightness(self.value)
        self.addCleanup(power.set_display_brightness, old_value)


class EnableRepowerdFakeSensors(fixtures.Fixture):
    """Fixture to enable fake sensors for repowerd."""

    def __init__(self):
        super().__init__()
        self.repowerd = RepowerdJob()

    def setUp(self):
        super().setUp()
        if not self.are_fake_sensors_enabled():
            self.enable_fake_sensors()
        self.addCleanup(self.repowerd.restart)

    def enable_fake_sensors(self):
        """Enable fake sensors on the repowerd process."""
        self.repowerd.restart(fake_sensors=True)
        validate_fifo_ready(self.repowerd.name)
        setup_sensors(self.repowerd.name)

    def are_fake_sensors_enabled(self):
        """Return True if fake sensors enabled on repowerd."""
        try:
            pid = get_process_id(self.repowerd.name)
        except (CalledProcessError, ValueError):
            return False
        return sensors.are_fake_sensors_enabled_for_process(pid)
