# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
from tempfile import mkstemp
from threading import Event

import psutil
from autopilot import (
    exceptions,
    input,
    logging as autopilot_logging,
    testcase
)
from autopilot.introspection import _object_registry as object_registry
from dbus.exceptions import DBusException
from retrying import retry

from ubuntu_system_tests.helpers import (
    autopilot,
    context,
    images,
    mir,
)
from ubuntu_system_tests.helpers import ofono
from ubuntu_system_tests.helpers import screen
from ubuntu_system_tests.helpers import sensors
from ubuntu_system_tests.helpers import wait_until
from ubuntu_system_tests.helpers.job_fixtures import JobEnvironmentVariable
from ubuntu_system_tests.helpers.unity8 import Unity8Job
from ubuntu_system_tests.helpers.location.fixture_setup import (
    SetLocationAccessRequests)
from ubuntu_system_tests.helpers.notifications import sim_pin_observer
from ubuntu_system_tests.helpers.processes import (
    ensure_all_apps_closed,
    is_testability_enabled_for_process,
    PidNotFoundError,
)
from ubuntu_system_tests.helpers.timeout import timeout
from ubuntu_system_tests.helpers.unity8 import get_dash
from ubuntu_system_tests.helpers.webapp import (
    DEFAULT_WEBVIEW_INSPECTOR_IP,
    DEFAULT_WEBVIEW_INSPECTOR_PORT,
)

# Any test name ending with this will leave device locked

TEST_ID_DEVICE_LOCKED = 'device_locked'

logger = logging.getLogger(__name__)

unity8 = Unity8Job()


class BaseUbuntuSystemTestCase(testcase.AutopilotTestCase):

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        unity8.ensure_unity_running_and_greeter_hidden()

    def setUp(self, allow_scopes_location=True, restart_unity=True,
              fake_sensors=False):
        super().setUp()
        self.useFixture(SetLocationAccessRequests(
            'unity8-dash', allow_scopes_location))
        self._reset_autopilot_registry()
        self._init_shared_context()
        self.useFixture(JobEnvironmentVariable(QT_LOAD_TESTABILITY='1'))
        self.setup_unity(restart_unity, fake_sensors)

    # Retry added until fix for bug 1427946 in QtMir is landed
    @retry(stop_max_attempt_number=5,
           wait_fixed=2000,
           retry_on_exception=lambda exception: (
               isinstance(exception, psutil.NoSuchProcess) or
               isinstance(exception, ValueError) or
               isinstance(exception, DBusException) or
               isinstance(exception, TimeoutError) or
               isinstance(exception, PidNotFoundError) or
               isinstance(exception, exceptions.ProcessSearchError) or
               isinstance(exception, exceptions.BackendException)))
    def setup_unity(self, restart_unity, fake_sensors):
        """Setup unity8 environment, restarting if any error is observed.
        :param restart_unity: Whether unity8 should be restarted.
        :param fake_sensors: Whether fake sensors should be enabled.
        """
        if restart_unity:
            self.unity_proxy = self.restart_unity(fake_sensors)
        else:
            # If unity8 is not restarted, we need to make sure
            # that testeability is enabled for the unity8 instance
            # which is already running, if it's not, then we have
            # no chance other than to restart unity8
            if not is_testability_enabled_for_process(
                    unity8.get_pid()):
                self.unity_proxy = self.restart_unity(fake_sensors)
            else:
                self.unity_proxy = self._get_unity8_proxy_object()
        self.addCleanup(ensure_all_apps_closed)
        context.shared.dash = get_dash()
        context.shared.dash.wait_for_dash_loaded()

    @property
    def shell(self):
        return self.unity_proxy.main_window

    @staticmethod
    def _init_shared_context():
        # The shared context is reset for each test
        context.shared.clear()
        context.shared.timers = []
        context.shared.retry_time = 10000

    @staticmethod
    def _get_unity8_proxy_object():
        from ubuntu_system_tests.helpers.unity8.shell import Unity8  # NOQA
        return autopilot.get_proxy_object(unity8.get_pid())

    @autopilot_logging.log_action(logger.info)
    def launch_unity(self, fake_sensors=False, dismiss_sim_unlock=True):
        """Launch Unity8 with testability enabled.

        :return: The Unity8 autopiliot custom proxy object.

        """
        if fake_sensors:
            self.useFixture(JobEnvironmentVariable(
                UBUNTU_PLATFORM_API_TEST_OVERRIDE='sensors')
            )
            self._start_unity_with_testability(no_wait=True)
            self._create_sensors()
        else:
            self._start_unity_with_testability()

        unity = self._get_unity8_proxy_object()
        context.shared.unity = unity
        self._wait_for_unity_ready(dismiss_sim_unlock)
        # Workaround to remove the UInput keyboard created by autopilot.
        # This stops the OSK from being displayed, so must remove it.
        self._remove_autopilot_keyboard()
        self._reset_keyboard()
        self._enable_osk_for_uitk()
        # Disable setting touch visualization due to bug:
        # https://bugs.launchpad.net/bugs/1602600
        # This should be re-enabled once an alternative method is provided.
        # self.set_touch_visualization()
        return unity

    def _wait_for_unity_ready(self, dismiss_sim_unlock):
        """Wait until unity is ready, dismissing sim pin entry notifications
        if required.
        :param dismiss_sim_unlock: Whether to dismiss sim pin screens.
        """
        if dismiss_sim_unlock:
            locked_sims = ofono.get_locked_sims()
            if locked_sims:
                # There are locked sims, so dismiss any pin entry screens.
                # The pin entry screen does not always show, depending on
                # whether the sims have been unlocked previously after boot.
                # So clear the screens from a different thread to avoid
                # blocking here if they don't show.
                completed = Event()
                pin_observer = sim_pin_observer.start_sim_pin_observer(
                    completed, locked_sims)
                # Wait until there are no notifications and greeter displayed.
                n = context.shared.unity.main_window.get_notifications_list()
                n.count.wait_for(0)
                context.shared.unity.wait_for_greeter_content()
                if not completed.is_set():
                    # There were no sim pin screens shown, but greeter is now
                    # ready, so stop the sim pin observer.
                    pin_observer.stop()
            else:
                context.shared.unity.wait_for_greeter_content()

    @staticmethod
    def _start_unity_with_testability(no_wait=False):
        args = ['QT_LOAD_TESTABILITY=1']
        if no_wait:
            args.append('--no-wait')
        unity8.start(*args)
        if no_wait:
            # just wait for job to be spawned rather than fully running
            unity8.wait_for_job_spawned()
        context.shared.unity_pid = unity8.get_pid()

    def _enable_osk_for_uitk(self):
        """Enable use of the OSK with uitk helpers."""
        os.environ['UITK_USE_MALIIT'] = '1'

    def _remove_autopilot_keyboard(self):
        """Remove the UInput simulated keyboard automatically created by
        autopilot which stops the OSK from appearing."""
        if self._kb:
            if not self._kb._device._device.device:
                self.assertTrue(
                    wait_until(self._find_keyboard),
                    'Could not find the UInput keyboard created by autopilot')
            try:
                self._kb._device._device.close()
            except OSError:
                pass
            self._kb = None

    def _find_keyboard(self):
        """Find the autopilot keyboard if it is available and save instace
        in the self._kb member variable.
        :return: True if keyboard found, False otherwise."""
        kb = self._kb._device._device._find_device()
        if kb:
            self._kb._device._device.device = kb
        return kb

    def _reset_autopilot_registry(self):
        """Reset autopilot's registry between tests to avoid CPO errors."""
        object_registry.patch_registry({})

    # Retry added until fix for bug 1511895 in ubuntu-keyboard is landed
    @retry(stop_max_attempt_number=3,
           retry_on_exception=lambda exception: (
               isinstance(exception, exceptions.ProcessSearchError) or
               isinstance(exception, exceptions.BackendException) or
               isinstance(exception, TimeoutError)))
    @timeout(30)
    def _reset_keyboard(self):
        """Reset the Autopilot connection to the OSK"""
        input.Keyboard.create('OSK')._keyboard.reset()

    # Retry added until fix for bug 1427946 in QtMir is landed
    @retry(stop_max_attempt_number=3,
           retry_on_exception=lambda exception: (
               isinstance(exception, FileNotFoundError) or
               isinstance(exception, AssertionError)))
    def _create_sensors(self):
        """Create fake sensors for testing"""
        sensors.validate_fifo_ready(unity8.name)
        sensors.setup_sensors(unity8.name)

    @autopilot_logging.log_action(logger.info)
    def restart_unity(self, fake_sensors=False):
        """
        Restart and unlock unity8

        :param fake_sensors: Whether to use the fake sensors
        :returns: The proxy object for the Unity8 shell

        """
        unity8.ensure_stopped()
        unity_proxy = self.launch_unity(fake_sensors=fake_sensors)
        if self.device_unlock_required():
            unity_proxy.unlock()

        return unity_proxy

    def set_touch_visualization(self):
        """Set touch visualization as required and disable on cleanup."""
        enable = os.getenv('TOUCH_VISUALIZATION') == 'true'
        screen.set_visualization(enable)
        if enable:
            self.addCleanup(screen.disable_touch_visualization)

    def device_unlock_required(self):
        """Return True if test requires device unlocking, False otherwise."""
        test_id = self.id().split('.')[-1]
        return not test_id.endswith(TEST_ID_DEVICE_LOCKED)

    def _get_screenshot_size(self, x, y, width, height):
        """ Calculate the screenshot size """
        unity = context.shared.unity
        if width and height:
            return width, height
        else:
            return unity.get_width(), unity.get_height()

    def get_screenshot_image(self, x=0, y=0, width=0, height=0):
        """ Take an screenshot and return a PIL Image file. The screenshot
        will be automatically deleted after the test execution. Passing the
        parameters for the rectangle coordinates, the screenshot captured will
        be just for the section selected
        :return: A PIL Image file which contains the screenshot
        """
        file_path = mkstemp()[1]
        self.addCleanup(os.remove, file_path)
        mir.get_rgba_screenshot(file_path, x, y, width, height)

        return images.get_image_from_rgba(
            file_path, self._get_screenshot_size(x, y, width, height))

    def setup_webapp_environment(self):
        """ Setup the environment for the webapps. This method has to be
        called before launch the first web app in the test.
        """
        self.useFixture(JobEnvironmentVariable(
            global_=True,
            UBUNTU_WEBVIEW_DEVTOOLS_HOST=DEFAULT_WEBVIEW_INSPECTOR_IP,
            UBUNTU_WEBVIEW_DEVTOOLS_PORT=str(DEFAULT_WEBVIEW_INSPECTOR_PORT)))
